import random
import numpy as np
import pandas as pd
import igraph as ig
from numpy.random import uniform

class GeneRegulatoryNetwork:
    
    def create_grn(self):
        # load the GRN and extract the transcription factor network
        if self.verbose:
            print ("Cleaning Biological Network...", flush=True)
            
        self.grn = self.get_grn()
        self.extract_tf_net()
        
        if self.tf_net.vcount() <= self.num_tfs:
            raise Exception("Number of Tfs is Larger than the GRN Provided...", flush=True)
        
        # sample tfs and e-genes
        if self.verbose:
            print ("Sampling TFs and E-Genes Network...", flush=True)
       
        self.sample_tfs()
        self.sample_egenes()
        
        # clean sampled network
        if self.verbose:
            print ("Pruning invalid edges...", flush=True)
            
        network = self.network_qc()
        self.tf_egene_net = self.rename_nodes(network, self.tfs, 'TF')
        self.tf_egene_net = self.rename_nodes(self.tf_egene_net, self.egenes, 'E-Gene')
    
    def get_grn(self):
        grn = pd.read_csv(self.grn_fp).set_index('row_names')
        edgelist = {'source':[], 'target':[], 'edge_weight':[]}

        for rowname, row in grn.iterrows():
            for colname in row.index:
                weight = row.loc[colname]

                if weight == 0:
                    continue

                edgelist['target'].append(colname)
                edgelist['source'].append(rowname)
                edgelist['edge_weight'].append(weight)

        edgelist = pd.DataFrame(edgelist)
        grn = ig.Graph.DataFrame(edgelist)
        return grn
    
    def extract_tf_net(self):
        tf_nodes = []
        nodes = [node.index for node in self.grn.vs]
        degrees = self.grn.degree(mode='out')

        for node_i, node in enumerate(nodes):
            if degrees[node_i] == 0:
                continue

            tf_nodes.append(node)

        tf_net = self.grn.induced_subgraph(tf_nodes)
        self.tf_net = self.clean_tf_net(tf_net)
        
    def sample_tfs(self):
        self.tfs = self.modular_sampling(self.tf_net, self.num_tfs)
        
    def sample_egenes(self):
        temp = self.grn
        
        # get neighbors from GRN of transcription factors
        neighbors = self.get_neighbors(temp, self.tfs, 'out')
        temp = temp.induced_subgraph(neighbors + self.tfs)
        
        # use page rank to calculate node weights
        page_rank = temp.pagerank(directed=True, weights=temp.es['edge_weight'], damping=0.1)
        page_rank = page_rank + np.random.uniform(size=len(page_rank))
        
        # remove TFs from Page Rank Results
        page_df = pd.DataFrame({'feature_id': temp.vs['name'], 'weight': page_rank})
        page_df = page_df.loc[~page_df['feature_id'].isin(self.tfs), ]
        
        if len(neighbors) < self.num_egenes:
            num_egenes = len(neighbors)
        else:
            num_egenes = self.num_egenes
        
        if len(neighbors) != len(page_rank):
            neighbors = neighbors[:len(page_rank)]
        
        # sample E-Genes from the neighbors
        self.egenes = random.choices(page_df['feature_id'].values,
                                     weights=page_df['weight'].values,
                                     k=num_egenes)
    
    def network_qc(self):
        network = self.remove_egenes_to_tf_edges()
        network = self.reattach_egenes(network)
        return network
    
    def clean_tf_net(self, tf_net):
        graphs = tf_net.decompose(mode="WEAK")
        graph_edges = [graph.ecount() for graph in graphs]
        most_edges = max(graph_edges)

        for index, edge_count in enumerate(graph_edges):

            if edge_count == most_edges:
                tf_net = graphs[index]

        return tf_net
    
    def remove_egenes_to_tf_edges(self):
        sampled_genes = self.tfs + self.egenes
        temp = self.grn.induced_subgraph(sampled_genes)
        df = self.get_edgelist_df(temp)
            
        df['valid_edge'] = False
        df.loc[(df.source.isin(sampled_genes)) | (df.target.isin(sampled_genes)), 'valid_edge'] = True
        df.loc[(~df.source.isin(self.tfs)) & (df.target.isin(self.tfs)), 'valid_edge'] = False
        
        df = df.loc[df.valid_edge, ]
        df = df.drop(columns=['valid_edge'])
        
        grn = ig.Graph.DataFrame(df)
        return grn
    
    def reattach_egenes(self, network):
        df = self.get_edgelist_df(network)
        missing_targets = [egene for egene in self.egenes if not egene in df['target'].values]
        
        if len(missing_targets) > 0:
            prob = self.network.degree(mode='out')
            prob = np.array(prob) / sum(prob)
            new_regs = random.choices(self.network.vs['name'], weights=prob, k=len(missing_targets))
            df = pd.concat([df, pd.DataFrame({'source': new_regs, 'target': missing_targets,
                                              'edge_weight': uniform(size=len(missing_targets))})])
        
        return ig.Graph.DataFrame(df)


import os
import random
import numpy as np
import pandas as pd
from tqdm import tqdm
import itertools as it

class GuideRNA:
    
    def generate_grnas(self, crispr_type=None, on_target=None, off_target=None):
        
        self.check_crispr_type(crispr_type)
        
        self.get_target_genes()
        self.check_target_scores(on_target, off_target)
        ngrnas_seq = range(self.ngrnas_per_target)
        ctrl_grnas = [f'{self.ctrl_label}-grna.{i + 1}' for i in ngrnas_seq]
        gene_grnas = [f'{gene}-grna.{i + 1}' for gene in self.target_genes for i in ngrnas_seq]
        
        grna_names = ctrl_grnas + gene_grnas
        self.ngrnas = len(grna_names)
        
        grna_meta = self.get_perturbed_genes(grna_names)
        grna_meta = self.sample_on_target(grna_meta)
        
        self.grna_meta = pd.DataFrame(grna_meta)
        
        if self.crispr_type == 'activation':
            self.grna_meta['on_target'] = (1 - self.grna_meta['on_target']) + 1
        
        self.create_multiplier_matrix()
    
    def get_target_genes(self):
        
        if len(self.target_genes) == 0:
            
            if self.perturb_tfs:
                tfs = list(self.feature_info.feature_id.loc[self.feature_info.is_tf])
                self.target_genes = self.target_genes + tfs
            
            if self.perturb_kinases:
                kinases = list(self.feature_info.feature_id.loc[self.feature_info.is_kinase])
                self.target_genes = self.target_genes + kinases
    
    def create_multiplier_matrix(self):
        
        if self.crispr_type == 'knockout':
            multiplier = self.create_knockout_meta()
            self.sim_meta = self.calc_sample_perc(multiplier, self.grna_meta, self.ctrl_label)
        else:
            multiplier = self.create_nonko_meta()
            self.sim_meta = pd.DataFrame({'sim_name': multiplier.index, 'grna': multiplier.index, 'sample_percent': 1})
        
        nrows = len(multiplier)
        self.calc_sims_per_device(nrows)
        self.multiplier = {"feature_id": [], "perturbation": [], "grna": [], "sim_i": []}
        
        for i in range(nrows):
            adjust_interval = 1 + i * self.nsims_per_condition
            grna = multiplier.index.values[i]
            perturb_vec = list(multiplier.iloc[i].T.values)
            
            for sim_i in range(self.nsims_per_condition):
                sim_i += adjust_interval
                feature = list(multiplier.columns.values + f"_{sim_i}") 
                self.multiplier["feature_id"] = self.multiplier["feature_id"] + feature
                self.multiplier["grna"] = self.multiplier["grna"] + [grna] * len(feature)
                self.multiplier["sim_i"] = self.multiplier["sim_i"] + [sim_i] * len(feature)
                self.multiplier["perturbation"] = self.multiplier["perturbation"] + perturb_vec
                
        self.multiplier = pd.DataFrame(self.multiplier)
        
    def get_perturbed_genes(self, grna_names):
        perturbed_genes = []
        
        for grna in tqdm(grna_names):
            sampled_genes = []
            
            if self.ctrl_label in grna:
                sampled_genes.append({'grna': grna, 'perturbed_gene': self.ctrl_label, 'on_target': 1, 'target': False})
            else:
                target_gene = grna.split('-grna')[0]
                sampled_genes = self.sample_off_target(target_gene, grna)
                sampled_genes.append({'grna': grna, 'perturbed_gene': target_gene, 'on_target': self.on_target, 'target': True})
            
            perturbed_genes = perturbed_genes + sampled_genes
        
        return perturbed_genes
    
    def sample_off_target(self, target_gene, grna):
        genes_to_add = []
        genes = list(self.feature_info.feature_id.values)
        off_target_genes = random.sample(genes, k=self.off_target)
        
        for index, gene in enumerate(off_target_genes):
            if gene == target_gene:
                continue
           
            genes_to_add.append({'grna': grna, 'perturbed_gene': gene, 'on_target': 0, 'target': False})
            
        if len(genes_to_add) != self.off_target:
            genes_to_add = self.sample_off_target(target_gene, grna)
        
        return genes_to_add
    
    def sample_on_target(self, grna_meta):
        probs = self.grna_library.probability
        on_target_scores = self.grna_library.on_target / 100
        
        for index, meta in enumerate(grna_meta):
            
            if not meta['target'] and meta['perturbed_gene'] != self.ctrl_label:
                meta['on_target'] = random.choices(on_target_scores, weights=probs, k=1)[0]
            
            grna_meta[index] = meta
        
        return grna_meta
    
    def create_nonko_meta(self):
        grnas = self.grna_meta.grna.unique()
        genes = self.feature_info['feature_id'].values
        multiplier = pd.DataFrame(data=1, index=grnas, columns=genes)

        for index, grna in enumerate(grnas):

            if not 'CTRL' in grna:
                grna_df = self.grna_meta.loc[self.grna_meta.grna == grna, ]
                
                for _, row in grna_df.iterrows():
                    col_index = None
                    
                    for i, col in enumerate(multiplier.columns):
                        if row.perturbed_gene == col:
                            col_index = i
                        
                    if col_index is None:
                        print (row.perturbed_gene)
                        
                    multiplier.iloc[index, col_index] = row.on_target
            
        return multiplier
    
    def create_knockout_meta(self):
        multiplier = []
        genes = self.feature_info['feature_id'].values
        
        for grna in self.grna_meta.grna.unique():
            if self.ctrl_label in grna:
                multi = pd.DataFrame(data=1, index=[grna], columns=genes)
                multiplier.append(multi)
                
            else:
                grna_df = self.grna_meta.loc[self.grna_meta.grna == grna, ]
                keys, values = zip(*{gene: [0, 1] for gene in grna_df.perturbed_gene.values}.items())
                ko_combo = pd.DataFrame([dict(zip(keys, v)) for v in  it.product(*values)])

                sims_label = self.create_sim_labels(grna, ko_combo)
                multi = pd.DataFrame(data=1, index=sims_label, columns=genes)
                multiplier.append(self.update_multiplier(multi, ko_combo))

        return pd.concat(multiplier)
    
    def create_sim_labels(self, grna, ko_combo):
        sims_label = []

        for row in range(len(ko_combo)):
            label = grna

            for col in range(len(ko_combo.columns)):
                gene = ko_combo.columns[col]
                prtb = ko_combo.iloc[row, col]

                if prtb == 0:
                    label = f'{label}_{gene}_PRT'
                else:
                    label = f'{label}_{gene}_NT'

            sims_label.append(label)

        return sims_label

    def update_multiplier(self, grna_multi, ko_combo):
        
        for col in ko_combo.columns:
            grna_multi[col] = ko_combo[col].values

        return grna_multi
    
    def calc_sample_perc(self, multiplier, grna_meta, ctrl_label):
        grnas = []
        sample_perc = []

        for row_i, rowname in enumerate(multiplier.index):

            grna_label = "_".join(rowname.split("_")[:2])
            grna = grna_meta.loc[grna_meta.grna == grna_label].copy()
            ngenes = grna.shape[0]
            grnas.append(grna_label)

            if ctrl_label in rowname:
                sample_perc.append(1)
                
            elif ngenes == 1:
                if 'PRT' in rowname:
                    if grna.on_target.values[0]== 0:
                        sample_perc.append(1)
                    else:
                        sample_perc.append(grna.on_target.values[0])
                    
                else:
                    if grna.on_target.values[0]== 0:
                        sample_perc.append(0)
                    else:
                        sample_perc.append(1 - grna.on_target.values[0])
                
            else:
                perc = self.multiple_genes_percentage(grna, multiplier, row_i)
                sample_perc.append(perc)

        sim_meta = pd.DataFrame({'sim_name': multiplier.index, 'grna': grnas, 'sample_percent': sample_perc})
        return sim_meta

    def multiple_genes_percentage(self, grna_meta, multiplier, sim_name):
        grna_meta['perturbed_gene'] = grna_meta['perturbed_gene'].values
        ko_combos = multiplier[list(grna_meta['perturbed_gene'].values)].iloc[sim_name]

        for col in ko_combos.index:
            value = ko_combos[col]
            gene_on_target = grna_meta.loc[grna_meta.perturbed_gene == col, 'on_target'].values[0]

            if value == 1:
                # probability of grna edit
                ko_combos[col] = 1 - gene_on_target
            else:
                # probability of no edit
                ko_combos[col] = gene_on_target

        return round(np.prod(ko_combos), 2)
    
    def calc_sims_per_device(self, nrows):
        nsims = 0
        total_sims = self.nsims_per_condition * nrows
        
        while nsims != total_sims:
            self.nsims_per_device = (self.nsims_per_condition * nrows) // self.nbatches
            nsims = self.nsims_per_device * self.nbatches

            if nsims != total_sims:
                self.nbatches += 1
    
    def check_crispr_type(self, crispr_type):
        
        try:
            crispr_type = crispr_type.lower()
            if crispr_type in ['interference', 'activation', 'knockout']:
                self.crispr_type = crispr_type
            else:
                raise Exception("crispr_type parameter must be either: activation, interference, or knockout...")
        except:
            if self.crispr_type is None:
                raise Exception("crispr_type parameter must be either: activation, interference, or knockout...")
                
    def check_target_scores(self, on_target, off_target):
        
        if (not on_target is None) and (type(on_target) is float):
            self.on_target = on_target
        
        if (not off_target is None) and (type(off_target) is int):
            self.off_target = off_target
        
        if self.off_target is None:
            self.off_target = random.choices(self.grna_library.off_target, weights=probs, k=1)[0]
        
        if self.on_target is None:
            self.on_target = random.choices(self.grna_library.on_target, weights=probs, k=1)[0]


import os
import random
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
from multiprocessing import Pool, RLock

class Sampling:
    
    def sample(self, ncells=10000, pop_fp=None, cache=True):
        print (f"Simulation: {self.network_name} Sampling Cells...", flush=True)
        cells_meta, gene_expr = self.sampling_cells(ncells)
        print (f"Simulation: {self.network_name} Sampling Molecules...", flush=True)
        gene_expr, lib_sizes = self.sampling_molecules(gene_expr, pop_fp)
        
        cells_meta = self.clean_cells_metadata(cells_meta, lib_sizes)
        cells_meta = cells_meta.reset_index(drop=True)
        
        if cache:
            print (f"Simulation: {self.network_name} Caching....", flush=True)
            cells_meta.to_csv(os.path.join(self.metadata_dir, 'cells_metadata.csv'), index=False)
            gene_expr.to_csv(os.path.join(self.metadata_dir, 'gene_expression.csv'), index=False)
        
        return cells_meta, gene_expr
        
    def sampling_cells(self, ncells=10000):
        ncells_from_sim = (self.perturb_time / self.update_interval)
        max_cells = int((self.sim_meta.shape[0] * self.nsims_per_condition) * ncells_from_sim)
        
        if ncells > max_cells:
            raise Exception(f"Simulation: {self.network_name} Number of cells requested is greater than the number of cells simulated. Sample fewer cells...")
        

        cells_meta = []
        cells = np.array([i for i in range(max_cells)])
        sims = (cells // ncells_from_sim).astype(int) + 1

        for sim, cell in tqdm(zip(sims, cells)):
            grna_label = self.multiplier['grna'].loc[self.multiplier.sim_i == sim].unique()[0]
            cells_meta.append({"cell_label": f"cell_{cell}", "sim_label": grna_label, "cell_i": cell,
                               "sim_i": sim, "fp": os.path.join(self.results_dir, f'simulation_{sim}.csv')})
        
        cells_meta = pd.DataFrame(cells_meta)
        
        if self.crispr_type == 'knockout':
            cells = self.ko_sample_cells(cells_meta, ncells)
        else:
            cells = list(random.choices(cells, k=ncells))
        
        cells_meta = cells_meta.iloc[cells]
        gene_expr = self.collapse_molecules(cells)
        
        return cells_meta, gene_expr
    
    def sampling_molecules(self, gene_expr, pop_fp=None):
        
        if pop_fp is None:
            pop = sc.read_loom(self.pop_fp)
        else:
            pop = sc.read_loom(pop_fp)
        
        realcounts = pop.X.toarray()
        cell_umi = pop.obs.total_counts.values
        
        lib_size = self.calc_library_size(cell_umi, gene_expr.values)
        simcounts_cpm = self.calc_cpm(realcounts, gene_expr.values)
        downsampled_counts = self.downsampling(simcounts_cpm, lib_size)
        gene_expr = pd.DataFrame(downsampled_counts, columns=gene_expr.columns)
        
        return gene_expr, lib_size
    
    def clean_cells_metadata(self, meta, lib_sizes):
        meta['lib_size'] = lib_sizes
        meta['grna'] = meta['sim_label'].apply(lambda x: "_".join(x.split('_')[0:2]))
        meta['target_gene'] = meta['sim_label'].apply(lambda x: x.split('-grna')[0])
        
        if self.crispr_type == 'knockout':
            meta['is_cell_perturbed'] = meta['sim_label'].apply(lambda x: x.split('_')[-1])
            meta.loc[meta.target_gene == self.ctrl_label, 'is_cell_perturbed'] = self.ctrl_label
        else:
            meta['is_cell_perturbed'] = 'PRT'
            meta.loc[meta.target_gene == self.ctrl_label, 'is_cell_perturbed'] = self.ctrl_label
        
        meta = meta.reset_index(drop=True)
        return meta
    
    def collapse_molecules(self, sampled_cells):
        df = pd.read_csv(os.path.join(self.results_dir, 'simulated_counts.csv'))
        df = df.sort_values(by=['sim_i'])
        df = df.reset_index(drop=True)
        df = df.iloc[sampled_cells]
        df = df.drop(columns=['sim_i'])
        
        if self.collapse_mrna:
            spec = pd.read_csv(os.path.join(self.species_vec_dir, 'batch_0.csv'))
            spec = spec.drop(columns=['species', 'state', 'sim_i'])
            spec = spec.loc[spec.molecule_type != 'protein', ]
            spec = spec.drop_duplicates(subset=['spec_name'])
            spec = spec.sort_values(by=['gene'])

            mrna = spec.loc[spec.molecule_type == 'mrna', 'spec_name'].values
            premrna = spec.loc[spec.molecule_type == 'premrna', 'spec_name'].values
            df = pd.DataFrame(df[mrna].values + df[premrna].values, columns=mrna)
        
        return df
        
    def calc_library_size(self, cell_umis, sim_counts):
        sim_counts_ls = np.sum(sim_counts, axis=1)
        
        if self.map_reference_ls:
            # sampling library
            sim_probs = np.random.uniform(size=len(sim_counts_ls))
            lib_size = np.around(np.quantile(cell_umis, sim_probs))
        else:
            lib_size = sim_counts_ls
        
        return lib_size
    
    def calc_cpm(self, realcount, sim_counts):
        realcount_ls = np.sum(realcount, axis=1).reshape(-1, 1)
        sim_counts_ls = np.sum(sim_counts, axis=1).reshape(-1, 1)
        
        realcount_cpm = realcount / realcount_ls
        sim_counts_cpm = sim_counts / sim_counts_ls
        
        if self.map_reference_cpm:
            # sort sim counts data via least to greatest
            sim_shape = sim_counts.shape
            realcount_cpm = realcount_cpm.flatten()
            sim_counts_cpm = sim_counts_cpm.flatten()
            
            realcount_cpm = realcount_cpm[realcount_cpm != 0]
            sim_probs = np.random.uniform(size=len(sim_counts_cpm))
            sim_counts_cpm = np.quantile(realcount_cpm, sim_probs).reshape(sim_shape)
            sim_counts_cpm = sim_counts_cpm / np.sum(sim_counts_cpm, axis=1).reshape(-1, 1)
        
        return sim_counts_cpm
    
    def downsampling(self, sim_counts_cpm, lib_sizes):
        
        for index, lib_size in tqdm(enumerate(lib_sizes)):
            gene_val = sim_counts_cpm[index, ]
            gene_expr = np.random.multinomial(lib_size, gene_val)
            sim_counts_cpm[index, ] = gene_expr
        
        return sim_counts_cpm
    
    def ko_sample_cells(self, cells_meta, ncells):
        sampled_cells = []
            
        for row_i in range(len(self.sim_meta)):
            row = self.sim_meta.iloc[row_i]
            sim_cells = cells_meta.loc[cells_meta.sim_label == row.sim_name, 'cell_i'].values
            
            if row.sample_percent != 0:
                n = int(ncells * row.sample_percent)
                sampled = random.choices(sim_cells, k=n)
                sampled_cells = sampled_cells + sampled
        
        return sampled_cells


import os
import math
import random
import numpy as np
import pandas as pd
from numpy.random import uniform
from scipy.stats import truncnorm

class Kinetics:
    """I am concerned about the basal levels/effects. Check what previous basal levels are like from the R code."""
        
    def initialize_kinetics(self):
        self.sample_rates()
        self.sample_interactions()
        self.calc_dissociation()
        
        grn = self.feature_network.loc[~self.feature_network.kinase_edge, ]
        kinases = self.feature_network.loc[self.feature_network.kinase_edge, ]
        
        # calculate for GRN
        self.calc_basal_activity(grn)
        self.calc_effects_sums(grn)
        self.calc_base_activity()
        
        # calculate for phosphorylation
        self.calc_basal_activity(kinases, "kinase_basal")
        self.calc_effects_sums(kinases, "kinase_effects_sums")
        self.calc_base_activity("kinase_base_activity", "kinase_basal", "kinase_effects_sums")
        
        self.feature_network = self.feature_network.drop(columns=['feature_id'])
        print ("Initialized Kinetics...")
        
        if self.cache_network:
            self.feature_info.to_csv(os.path.join(self.metadata_dir, 'feature_info.csv'), index=False)
            self.feature_network.to_csv(os.path.join(self.metadata_dir, 'feature_network.csv'), index=False)
    
    def sample_rates(self):
        nrows = len(self.feature_info)
        
        self.feature_info['effects_sums'] = 0
        self.feature_info['independence'] = 1
        self.feature_info['base_activity'] = 0
        self.feature_info['kinase_effects_sums'] = 0
        self.feature_info['kinase_base_activity'] = 0
        self.feature_info['phosphorylation_rate'] = 0
        self.feature_info['dephosphorylation_rate'] = 0
        self.feature_info['transcription_rate'] = uniform(size=nrows, high=20, low=10)
        self.feature_info['translation_rate'] = uniform(size=nrows, high=150, low=100)
        self.feature_info['mrna_halflife'] = uniform(size=nrows, high=5, low=2.5)
        self.feature_info['protein_halflife'] = uniform(size=nrows, high=10, low=5)
        self.feature_info['splicing_rate'] = math.log(2) / 2
        self.feature_info['mrna_decay_rate'] = math.log(2) / self.feature_info['mrna_halflife']
        self.feature_info['protein_decay_rate'] = math.log(2) / self.feature_info['protein_halflife']
        self.feature_info.loc[self.feature_info.is_phosphorylated, 'phosphorylation_rate'] = uniform(size=len(self.phosphorylated), high=20, low=10)
        self.feature_info.loc[self.feature_info.is_phosphorylated, 'dephosphorylation_rate'] = uniform(size=len(self.phosphorylated), high=20, low=10)
        
    def sample_interactions(self):
        nrows = len(self.feature_network)
        fnet = self.feature_network.copy(deep=True)
        
        fnet['basal'] = 0
        fnet['effect'] = random.choices([-1, 1], weights=[0.25, 0.75], k=nrows)
        fnet['strength'] = 10 ** uniform(size=nrows, high=math.log10(100), low=math.log10(1))
        fnet['hill'] = truncnorm.rvs(1, 10, loc=2, scale=2, size=nrows)
        
        self.feature_network = fnet
        
    def calc_dissociation(self):
        remove = ["max_premrna", "max_mrna", "max_protein", "dissociation", "k", "max_protein"]
        
        finfo, fnet = self.feature_info, self.feature_network
        fnet = fnet[[col for col in fnet.columns if not col in remove]]
        finfo = finfo[[col for col in finfo.columns if not col in remove]]
        
        finfo["max_premrna"] = finfo["transcription_rate"] / (finfo["mrna_decay_rate"] + finfo["splicing_rate"])
        finfo["max_mrna"] = finfo["splicing_rate"] / finfo["mrna_decay_rate"] * finfo["max_premrna"]
        finfo["max_protein"] = finfo["translation_rate"] / finfo["protein_decay_rate"] * finfo["max_mrna"] 
        
        fnet = fnet.merge(finfo[['feature_id', 'max_protein']], left_on='from', right_on='feature_id', how='left')
        fnet['dissociation'] = fnet['max_protein'] / 2
        fnet['feature_id'] = fnet['to']
        
        self.feature_info, self.feature_network = finfo, fnet
        
    def calc_basal_activity(self, network, basal_col="basal"):
        basal_df = {"feature_id": [], basal_col: []}

        for index, group in network[['feature_id', 'effect']].groupby(['feature_id']):
            effects = group.effect.unique()
            basal_df["feature_id"].append(index)

            if len(effects) == 2:
                basal_df[basal_col].append(0.5)
            elif effects[0] == -1:
                basal_df[basal_col].append(1)
            elif effects[0] == 1:
                basal_df[basal_col].append(0.0001)
            else:
                basal_df[basal_col].append(np.nan)

        basal_df = pd.DataFrame(basal_df)
        self.feature_info = self.feature_info.merge(basal_df, on='feature_id', how='left')
        self.feature_info.loc[self.feature_info[basal_col].isna(), basal_col] = 1
        # self.feature_network = self.feature_network.drop(columns=['feature_id'])
    
    def calc_effects_sums(self, network, effects_col="effects_sums"):
        # calculating effects sums
        info = self.feature_info
        info.index = info.feature_id

        for feature, group in network.groupby(['to']):
            info.at[feature, effects_col] = sum(group.effect.values)

        info.reset_index(inplace=True, drop=True)
        self.feature_info = info
    
    def calc_base_activity(self, base_active_col="base_activity", basal_col="basal", effects_col="effects_sums"):
        info = self.feature_info
        above_zero = info[effects_col] > 0
        equal_or_below_zero = info[effects_col] <= 0

        info.loc[equal_or_below_zero, base_active_col] = info.loc[equal_or_below_zero, basal_col]
        info.loc[above_zero, base_active_col] = info.loc[above_zero, basal_col] - info.loc[above_zero, 'independence'] ** info.loc[above_zero, effects_col]
        info.loc[info[base_active_col].isna(), base_active_col] = 1
        self.feature_info = info